#!/usr/bin/python3
'Utility to perform commands on PDF files.'
# Author: Mark Blakeney, Dec 2019.

import sys
import argparse
import importlib
from pathlib import Path

def import_path(path):
    'Import given module path'
    modname = path.stem.replace('-', '_')
    spec = importlib.util.spec_from_loader(modname,
            importlib.machinery.SourceFileLoader(modname, str(path)))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module

def main():
    'Main code'
    mainparser = argparse.ArgumentParser(description=__doc__)
    subparser = mainparser.add_subparsers(help='Available commands:',
            dest='func')

    # Iterate over the commands to set up their parsers
    for modfile in (Path(__file__).parent / 'commands').glob('[!_]*.py'):
        mod = import_path(modfile)
        docstr = mod.__doc__.strip() if mod.__doc__ else None
        parser = subparser.add_parser(modfile.stem, description=docstr,
                formatter_class=argparse.RawDescriptionHelpFormatter,
                help=docstr)

        # Add reference to mainparser in case subcommand needs it
        parser._mainparser = mainparser

        if hasattr(mod, 'init'):
            mod.init(parser)
        if not hasattr(mod, 'main'):
            mainparser.error(f'"{modfile.stem}" command must define a main()')

        parser.set_defaults(func=mod.main, parser=parser)

    args = mainparser.parse_args()

    if not args.func:
        mainparser.error('Must specify a command')

    # Run the command that the user specified
    args.func(args)

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/python3
'Output list of page labels/numbers, or total number of pages.'
# Author: Mark Blakeney, Dec 2019.
import sys
import argparse
import PyPDF2

def init(parser):
    'Called to add arguments to parser at init'
    parser.add_argument('file', help='PDF file', type=argparse.FileType('rb'))
    parser.add_argument('-c', '--count', action='store_true',
            help='just print the total count of pages')

def int_to_page_alpha(pageno, base):
    (div, mod) = divmod(pageno - 1, 26)
    c = chr(mod + ord(base))
    return c * (div + 1)

def int_to_roman(input):
    if not isinstance(input, type(1)):
        raise TypeError("expected integer, got %s" % type(input))
    if not 0 < input < 4000:
        raise ValueError("Argument must be between 1 and 3999")
    ints = (1000, 900, 500, 400, 100, 90, 50, 40, 10, 9, 5, 4, 1)
    nums = ('M', 'CM', 'D', 'CD', 'C', 'XC', 'L', 'XL', 'X', 'IX', 'V',
            'IV', 'I')
    result = []
    for i in range(len(ints)):
        count = int(input / ints[i])
        result.append(nums[i] * count)
        input -= ints[i] * count
    return ''.join(result)

def getpages(pdf):
    'Generate a list of page labels'
    try:
        nums = pdf.trailer["/Root"]["/PageLabels"]["/Nums"]
    except Exception:
        return range(1, pdf.getNumPages() + 1)

    # At this point we have either the object or the list.
    # Make it a list.
    if not isinstance(nums, list):
        nums = [nums]

    labels = []
    style = None
    # default
    style = '/D'
    prefix = ''
    next_pageno = 1
    for i in range(0, pdf.getNumPages()):
        if len(nums) > 0 and i >= nums[0]:
            nums.pop(0)  # discard index
            pnle = nums.pop(0)
            style = pnle.get('/S', '/D')
            prefix = pnle.get('/P', '')
            next_pageno = pnle.get('/St', 1)
        if style == '/D':
            pageno_str = str(next_pageno)
        elif style == '/A':
            pageno_str = int_to_page_alpha(next_pageno, 'A')
        elif style == '/a':
            pageno_str = int_to_page_alpha(next_pageno, 'a')
        elif style == '/R':
            pageno_str = int_to_roman(next_pageno)
        elif style == '/r':
            pageno_str = int_to_roman(next_pageno).lower()
        else:
            raise 'Malformed PDF: unkown page numbering style ' + style
        labels.append(prefix or pageno_str)
        next_pageno += 1

    return labels

def main(args):
    'Called to action this command'
    pdf = PyPDF2.PdfFileReader(args.file)
    if args.count:
        print(pdf.getNumPages())
        sys.exit()
    for page in getpages(pdf):
        print(page)



#!/usr/bin/python3
'Show PDF document information.'
# Author: Mark Blakeney, Dec 2019.
import argparse
import PyPDF2

def init(parser):
    'Called to add arguments to parser at init'
    parser.add_argument('file', help='PDF file', type=argparse.FileType('rb'))

def main(args):
    'Called to action this command'
    inf = PyPDF2.PdfFileReader(args.file).getDocumentInfo()
    for key, val in inf.items():
        if key.startswith('/'):
            key = key[1:]
        print(f'{key}: {val}')



#!/usr/bin/python3
'Concaternate selected pages of one or more PDF files into a single file.'
# Author: Mark Blakeney, Dec 2019.
import os
import re
import sys
import string
import argparse
import PyPDF2
from pathlib import Path

# flake8: noqa: E122
HELP = string.Template(\
'''
Arguments are specified as a sequence of alternating file names and page
ranges:

  file1 pagerange1 file2 pagerange2 ..

Page ranges refer to the previously specified file and are specified in
Python "slice" notation, assuming each page is like an element in a
sequence, starting at 0.

For example, concatenate all of head.pdf, all but page seven of
content.pdf, and the last page of tail.pdf, producing output.pdf:

  $prog cat -o output.pdf head.pdf content.pdf :6 7: tail.pdf -1

A file not followed by a page range means all pages of that file, for
example:

  $prog cat chapter*.pdf >book.pdf

You can also assign a single upper case "alias" name to a file at first
use and then use that alias later as shorthand. For example, output the
first page from document1.pdf, the first page of document2.pdf, then the
remaining pages from document1.pdf and document2.pdf:

  $prog cat -o output.pdf A=document1.pdf 0 B=document2.pdf 0 A 1: B 1:

Page range expression examples (remember, page indices start with zero):
      :     all pages.                   -1    last page.
      22    just the 23rd page.          :-1   all but the last page.
      0:3   the first three pages.       -2    second-to-last page.
      :3    the first three pages.       -2:   last two pages.
      5:    from the sixth page onward.  -3:-1 third & second to last.
  The third, "stride" or "step" number is also recognized.
      ::2       0 2 4 ... to the end.    3:0:-1    3 2 1 but not 0.
      1:10:2    1 3 5 7 9                2::-1     2 1 0.
      ::-1      all pages in reverse order.
''')

def init(parser):
    'Called to add arguments to parser at init'
    parser.add_argument('-o', '--outfile', default='-',
            help='output file, default = stdout')
    # Need to parse remaining args ourself because of ranges starting with '-'
    parser.add_argument('fileranges', nargs=argparse.REMAINDER,
            help='Sequence of alternating file names and page ranges')

    parser.epilog = HELP.safe_substitute(prog=Path(sys.argv[0]).stem)

def main(args):
    'Called to action this command'
    if not args.fileranges:
        args.parser.error('Must specify at least one input file')

    aliases = {}
    merger = PyPDF2.PdfFileMerger()
    for fname, pages in PyPDF2.parse_filename_page_ranges(args.fileranges):
        if re.search('^[A-Z]=.', fname):
            alias = fname[0]
            fname = fname[2:]
            aliases[alias] = fname
        else:
            fname = aliases.get(fname, fname)

        if not Path(fname).exists():
            sys.exit(f'File "{fname}" does not exist.')

        merger.append(fname, pages=pages)

    merger.write(os.fdopen(sys.stdout.fileno(), 'wb') if args.outfile ==
            '-' else args.outfile)
